@@ -14,9 +14,10 @@
 #include <QImageWriter>
 #include <QMimeData>
 #include <QPointer>
-#include <QtWaylandClient/QWaylandClientExtension>
+#include <QWaylandClientExtension>
+#include <QWindow>
+#include <QtWaylandClientVersion>
 #include <qpa/qplatformnativeinterface.h>
-#include <qtwaylandclientversion.h>
 
 #include <errno.h>
 #include <poll.h>
@@ -24,6 +25,7 @@
 #include <string.h>
 #include <unistd.h>
 
+#include "qwayland-wayland.h"
 #include "qwayland-wlr-data-control-unstable-v1.h"
 
 static inline QString applicationQtXImageLiteral()
@@ -72,13 +74,7 @@ public:
 
     void instantiate()
     {
-#if QTWAYLANDCLIENT_VERSION >= QT_VERSION_CHECK(6, 2, 0)
         initialize();
-#else
-        // QWaylandClientExtensionTemplate invokes this with a QueuedConnection but we want shortcuts
-        // to be have access to data_control immediately.
-        QMetaObject::invokeMethod(this, "addRegistryListener");
-#endif
     }
 
     ~DataControlDeviceManager()
@@ -154,11 +150,7 @@ protected:
         m_receivedFormats << mime_type;
     }
 
-#if QT_VERSION >= QT_VERSION_CHECK(6, 0, 0)
     QVariant retrieveData(const QString &mimeType, QMetaType type) const override;
-#else
-    QVariant retrieveData(const QString &mimeType, QVariant::Type type) const override;
-#endif
 
 private:
     /** reads data from a file descriptor with a timeout of 1 second
@@ -168,11 +160,7 @@ private:
     QStringList m_receivedFormats;
 };
 
-#if QT_VERSION >= QT_VERSION_CHECK(6, 0, 0)
 QVariant DataControlOffer::retrieveData(const QString &mimeType, QMetaType type) const
-#else
-QVariant DataControlOffer::retrieveData(const QString &mimeType, QVariant::Type type) const
-#endif
 {
     Q_UNUSED(type);
 
@@ -286,7 +274,11 @@ public:
 
     QMimeData *mimeData()
     {
-        return m_mimeData;
+        return m_mimeData.get();
+    }
+    std::unique_ptr<QMimeData> releaseMimeData()
+    {
+        return std::move(m_mimeData);
     }
 
 Q_SIGNALS:
@@ -297,7 +289,7 @@ protected:
     void zwlr_data_control_source_v1_cancelled() override;
 
 private:
-    QMimeData *m_mimeData;
+    std::unique_ptr<QMimeData> m_mimeData;
 };
 
 DataControlSource::DataControlSource(struct ::zwlr_data_control_source_v1 *id, QMimeData *mimeData)
@@ -448,6 +440,7 @@ private:
 
     std::unique_ptr<DataControlSource> m_primarySelection; // selection set locally
     std::unique_ptr<DataControlOffer> m_receivedPrimarySelection; // latest selection set from externally to here
+    friend WaylandClipboard;
 };
 
 void DataControlDevice::setSelection(std::unique_ptr<DataControlSource> selection)
@@ -455,7 +448,6 @@ void DataControlDevice::setSelection(std::unique_ptr<DataControlSource> selectio
     m_selection = std::move(selection);
     connect(m_selection.get(), &DataControlSource::cancelled, this, [this]() {
         m_selection.reset();
-        Q_EMIT selectionChanged();
     });
     set_selection(m_selection->object());
     Q_EMIT selectionChanged();
@@ -466,7 +458,6 @@ void DataControlDevice::setPrimarySelection(std::unique_ptr<DataControlSource> s
     m_primarySelection = std::move(selection);
     connect(m_primarySelection.get(), &DataControlSource::cancelled, this, [this]() {
         m_primarySelection.reset();
-        Q_EMIT primarySelectionChanged();
     });
 
     if (zwlr_data_control_device_v1_get_version(object()) >= ZWLR_DATA_CONTROL_DEVICE_V1_SET_PRIMARY_SELECTION_SINCE_VERSION) {
@@ -474,9 +465,78 @@ void DataControlDevice::setPrimarySelection(std::unique_ptr<DataControlSource> s
         Q_EMIT primarySelectionChanged();
     }
 }
+class Keyboard;
+// We are binding to Seat/Keyboard manually because we want to react to gaining focus but inside Qt the events are Qt and arrive to late
+class KeyboardFocusWatcher : public QWaylandClientExtensionTemplate<KeyboardFocusWatcher>, public QtWayland::wl_seat
+{
+    Q_OBJECT
+public:
+    KeyboardFocusWatcher()
+        : QWaylandClientExtensionTemplate(5)
+    {
+        initialize();
+        auto native = qGuiApp->platformNativeInterface();
+        auto display = static_cast<struct ::wl_display *>(native->nativeResourceForIntegration("wl_display"));
+        // so we get capabilities
+        wl_display_roundtrip(display);
+    }
+    ~KeyboardFocusWatcher() override
+    {
+        if (isActive()) {
+            release();
+        }
+    }
+    void seat_capabilities(uint32_t capabilities) override
+    {
+        const bool hasKeyboard = capabilities & capability_keyboard;
+        if (hasKeyboard && !m_keyboard) {
+            m_keyboard = std::make_unique<Keyboard>(get_keyboard(), *this);
+        } else if (!hasKeyboard && m_keyboard) {
+            m_keyboard.reset();
+        }
+    }
+    bool hasFocus() const
+    {
+        return m_focus;
+    }
+Q_SIGNALS:
+    void keyboardEntered();
+
+private:
+    friend Keyboard;
+    bool m_focus = false;
+    std::unique_ptr<Keyboard> m_keyboard;
+};
+
+class Keyboard : public QtWayland::wl_keyboard
+{
+public:
+    Keyboard(::wl_keyboard *keyboard, KeyboardFocusWatcher &seat)
+        : wl_keyboard(keyboard)
+        , m_seat(seat)
+    {
+    }
+    ~Keyboard()
+    {
+        release();
+    }
+
+private:
+    void keyboard_enter([[maybe_unused]] uint32_t serial, [[maybe_unused]] wl_surface *surface, [[maybe_unused]] wl_array *keys) override
+    {
+        m_seat.m_focus = true;
+        Q_EMIT m_seat.keyboardEntered();
+    }
+    void keyboard_leave([[maybe_unused]] uint32_t serial, [[maybe_unused]] wl_surface *surface) override
+    {
+        m_seat.m_focus = false;
+    }
+    KeyboardFocusWatcher &m_seat;
+};
 
 WaylandClipboard::WaylandClipboard(QObject *parent)
     : KSystemClipboard(parent)
+    , m_keyboardFocusWatcher(new KeyboardFocusWatcher)
     , m_manager(new DataControlDeviceManager)
 {
     connect(m_manager.get(), &DataControlDeviceManager::activeChanged, this, [this]() {
@@ -489,7 +549,6 @@ WaylandClipboard::WaylandClipboard(QObject *parent)
             if (!seat) {
                 return;
             }
-
             m_device.reset(new DataControlDevice(m_manager->get_data_device(seat)));
 
             connect(m_device.get(), &DataControlDevice::receivedSelectionChanged, this, [this]() {
@@ -532,6 +591,23 @@ void WaylandClipboard::setMimeData(QMimeData *mime, QClipboard::Mode mode)
     if (!m_device) {
         return;
     }
+
+    // roundtrip to have accurate focus state when losing focus but setting mime data before processing wayland events.
+    auto native = qGuiApp->platformNativeInterface();
+    auto display = static_cast<struct ::wl_display *>(native->nativeResourceForIntegration("wl_display"));
+    wl_display_roundtrip(display);
+
+    // If the application is focused, use the normal mechanism so a future paste will not deadlock itselfs
+    if (m_keyboardFocusWatcher->hasFocus()) {
+        QGuiApplication::clipboard()->setMimeData(mime, mode);
+        // if we short-circuit the wlr_data_device, when we receive the data
+        // we cannot identify ourselves as the owner
+        // because of that we act like it's a synchronous action to not confuse klipper.
+        wl_display_roundtrip(display);
+        return;
+    }
+    // If not, set the clipboard once the app receives focus to avoid the deadlock
+    connect(m_keyboardFocusWatcher.get(), &KeyboardFocusWatcher::keyboardEntered, this, &WaylandClipboard::gainedFocus, Qt::UniqueConnection);
     auto source = std::make_unique<DataControlSource>(m_manager->create_data_source(), mime);
     if (mode == QClipboard::Clipboard) {
         m_device->setSelection(std::move(source));
@@ -540,6 +616,22 @@ void WaylandClipboard::setMimeData(QMimeData *mime, QClipboard::Mode mode)
     }
 }
 
+void WaylandClipboard::gainedFocus()
+{
+    disconnect(m_keyboardFocusWatcher.get(), &KeyboardFocusWatcher::keyboardEntered, this, nullptr);
+    // QClipboard takes ownership of the QMimeData so we need to transfer and unset our selections
+    if (auto &selection = m_device->m_selection) {
+        std::unique_ptr<QMimeData> data = selection->releaseMimeData();
+        WaylandClipboard::clear(QClipboard::Clipboard);
+        QGuiApplication::clipboard()->setMimeData(data.release(), QClipboard::Clipboard);
+    }
+    if (auto &primarySelection = m_device->m_primarySelection) {
+        std::unique_ptr<QMimeData> data = primarySelection->releaseMimeData();
+        WaylandClipboard::clear(QClipboard::Selection);
+        QGuiApplication::clipboard()->setMimeData(data.release(), QClipboard::Selection);
+    }
+}
+
 void WaylandClipboard::clear(QClipboard::Mode mode)
 {
     if (!m_device) {
@@ -547,9 +639,11 @@ void WaylandClipboard::clear(QClipboard::Mode mode)
     }
     if (mode == QClipboard::Clipboard) {
         m_device->set_selection(nullptr);
+        m_device->m_selection.reset();
     } else if (mode == QClipboard::Selection) {
         if (zwlr_data_control_device_v1_get_version(m_device->object()) >= ZWLR_DATA_CONTROL_DEVICE_V1_SET_PRIMARY_SELECTION_SINCE_VERSION) {
             m_device->set_primary_selection(nullptr);
+            m_device->m_primarySelection.reset();
         }
     }
 }
