@@ -12,20 +12,18 @@
 #include <QAbstractNativeEventFilter>
 class KModifierKeyInfoProviderXcb : public KModifierKeyInfoProvider, public QAbstractNativeEventFilter
 {
-Q_OBJECT
-Q_PLUGIN_METADATA(IID "org.kde.kguiaddons.KModifierKeyInfoProvider.XCB")
+    Q_OBJECT
 public:
     KModifierKeyInfoProviderXcb();
-    ~KModifierKeyInfoProviderXcb();
+    ~KModifierKeyInfoProviderXcb() override;
 
     bool setKeyLatched(Qt::Key key, bool latched) override;
     bool setKeyLocked(Qt::Key key, bool locked) override;
 
-    bool nativeEventFilter(const QByteArray &eventType, void *message, long int *result) override;
+    bool nativeEventFilter(const QByteArray &eventType, void *message, qintptr *) override;
 
     void xkbUpdateModifierMapping();
-    void xkbModifierStateChanged(unsigned char mods, unsigned char latched_mods,
-                                 unsigned char locked_mods);
+    void xkbModifierStateChanged(unsigned char mods, unsigned char latched_mods, unsigned char locked_mods);
     void xkbButtonStateChanged(unsigned short ptr_buttons);
 
 private:
